@@ -127,6 +127,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
+#include <ctype.h>
 #include <sys/stat.h>
 #include <dirent.h>
 #include <sys/types.h>
@@ -174,7 +175,7 @@ static char *URLbasename = NULL;
    free will be exit() anyway.  Hack to fix a malloc bug that I couldn't
    track down (michael@actrix.gen.nz).
    */
-#define free(x)
+/* #define free(x) */
 
 static char
 *stralloc(int len)
@@ -1234,7 +1235,7 @@ static char
 	next_row(currow);
 	curfield=layout->first;
 	i=0;
-	while (!finished) {
+	while (!finished && *c) {
 		/* search item */
 		h=c;
 		if ((*c=='_' || *c=='=') && (c[1]==itemsep || c[1]=='\n')) {
@@ -2580,7 +2581,7 @@ static char
 					out_html("<HTML><HEAD><TITLE>"
 						 "Manpage of ");
 					out_html(wordlist[0]);
-					out_html("</TITLE>\n</HEAD><BODY BGCOLOR=\"#ffffff\">"
+					out_html("</TITLE>\n</HEAD><BODY>"
 						 "\n<H1>");
 					out_html(wordlist[0]);
 					out_html("</H1>\nSection: ");
@@ -3607,12 +3608,22 @@ static char
 				} else if (*h>31 && *h<127)
 					intbuff[ibp++]=*h;
 				else if (((unsigned char)(*h))>127) {
+#if 0
 					intbuff[ibp++]='&';
 					intbuff[ibp++]='#';
 					intbuff[ibp++]='0'+((unsigned char)(*h))/100;
 					intbuff[ibp++]='0'+(((unsigned char)(*h))%100)/10;
 					intbuff[ibp++]='0'+((unsigned char)(*h))%10;
 					intbuff[ibp++]=';';
+#endif
+					/* Do not convert this, as it may be
+					 * part of a multibyte sequence.  The
+					 * HTML viewer should hopefully know
+					 * what to do with the encoding.  Hmm,
+					 * we could probably merge this case
+					 * with the one above.
+					 */
+					intbuff[ibp++] = (guchar) *h;
 				}
 				curpos++;
 				break;
@@ -3671,12 +3682,22 @@ int
 main(int argc, char **argv)
 {
 	char *t=NULL;
-	int i;
+	int i,len;
 	char *buf;
 	char *h = '\0';
 	STRDEF *stdf;
 	char *infile = NULL;
 
+#if defined HAVE_MAN_WHEREIS
+#define MAN_CAN_FIND_SOURCE
+   const char *man_source_locator = "man -w %s";
+   const char *man_source_locator_with_section = "man -w %c %s";
+#elif defined HAVE_MAN_DEBUG
+#define MAN_CAN_FIND_SOURCE
+   const char *man_source_locator = "man -d %s | sed -n 's,^[ \t]*unformatted[ \t]*=[ \t]*,,p'";
+   const char *man_source_locator_with_section = "man -d -s %c %s | sed -n 's,^[ \t]*unformatted[ \t]*=[ \t]*,,p'";
+#endif
+
 	/* see if they gave us a basename for the URL references */
 
 	for(i = 1; i < argc; i++)
@@ -3722,18 +3743,65 @@ main(int argc, char **argv)
 
 		    *ctmp = '\0';
 
-		    g_snprintf(cmdline, sizeof(cmdline), "man -w %c %s", section, infile);
+#ifdef MAN_CAN_FIND_SOURCE
+		    g_snprintf(cmdline, sizeof(cmdline), man_source_locator_with_section, section, infile);
+#else
+		{
+                  printf("<HTML><HEAD><TITLE>Document not found</TITLE>\n"
+                         "</HEAD><BODY>"
+                         "I don't know how to find the document \"%s\" "
+                         "on your system.\n"
+                         "</BODY></HTML>\n", infile);
+
+                  return 4;
+                }
+#endif
 		  }
 		else
-		  g_snprintf(cmdline, sizeof(cmdline), "man -w %s", infile);
+#ifdef MAN_CAN_FIND_SOURCE
+		  g_snprintf(cmdline, sizeof(cmdline), man_source_locator, infile);
+#else
+		{
+                  printf("<HTML><HEAD><TITLE>Document not found</TITLE>\n"
+                         "</HEAD><BODY>"
+                         "I don't know how to find the document \"%s\" "
+                         "on your system.\n"
+                         "</BODY></HTML>\n", infile);
 
+                  return 4;
+                }
+#endif
 		fh = popen(cmdline, "r");
 		fgets(output, sizeof(output), fh);
 		pclose(fh);
-
+		
+		/* 
+		 * If man page has been cached 'man -w cmdname' will
+		 * return both the location of the manpage and the
+		 * location of the cached cat version separated by one
+		 * or more spaces.  Need to trim second entry and/or
+		 * any trailing spaces after first entry.
+		 *
+		 * However man pages can theoretically have spaces in
+		 * them.  we would use man -W, but that seems to not
+		 * be portable.
+		 *
+		 * So, look for ' /' which is probably going to work
+		 * right.
+		 */
+		
+		len = strlen(output);
+		for(i = 0; i < len; i++)
+		{
+			if (isspace((unsigned char)output[i] && (output[i+1] == '/')))
+				output[i] = '\0';
+		}
+			
 		i = strlen(output) - 1;
-		while(isspace((unsigned char)output[i])) output[i--] = '\0';
+		while (isspace((unsigned char)output[i])) 
+			output[i--] = '\0';
 
+		
 		if (output[0]) {
 #ifdef HAVE_LIBBZ2
 		  if(strstr(output,".bz2"))
